#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a Cairo plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;;; Scheme Graphics Operations to a Cairo Image Surface.
;;; package: (cairo graphics)

(define-structure (cairo-graphics
		   (constructor %make-cairo-graphics
				(width height surface context)))

  ;; Cairo Surface -- |cairo_surface_t| alien.
  surface width height

  ;; Cairo Context -- |cairo_t| alien.
  context

  ;; Bounds to which context is clipped, scaled, translated...
  (limits '(-1. -1. 1. 1.))
  (scale 1.)

  (fgcolor #f)
  (fgcolor-name #f)
  (bgcolor #f)
  (bgcolor-name #f))

(define (make-cairo-graphics width height)
  (let ((factor (->flonum (/ (min (-1+ width) (-1+ height)) 2))))
    (if (not (flo:positive? factor))
	(error "Invalid width x height:" width height))
    (let* ((surface (cairo-image-surface-create width height))
	   (cr (cairo-create surface))
	   (graphics (%make-cairo-graphics width height surface cr)))
      (cairo-save cr)
      (cairo-scale cr factor (flo:negate factor))
      (set-cairo-graphics-scale! graphics factor)
      (cairo-set-line-width cr (flo:/ 1.0 factor))
      (cairo-set-font-matrix cr (let ((k (flo:/ 10.0 factor)))
				  (cairo-matrix k 0. 0.
						0. (flo:negate k) 0.)))
      (cairo-translate cr 1.0 -1.0)
      (cairo-set-source-color cr white)
      (cairo-paint cr)
      (cairo-set-source-color cr black)
      (set-cairo-graphics-bgcolor-name! graphics "white")
      (set-cairo-graphics-bgcolor! graphics white)
      (set-cairo-graphics-fgcolor-name! graphics "black")
      (set-cairo-graphics-fgcolor! graphics black)
      graphics)))

(define (cairo-graphics/available?) #t)

(define (cairo-graphics/open make-device #!optional width height)
  (let ((width (if (default-object? width) 512 width))
	(height (if (default-object? height) 384 height)))
    (guarantee positive-fixnum? width 'cairo-graphics/open)
    (guarantee positive-fixnum? height 'cairo-graphics/open)
    (make-device (make-cairo-graphics width height))))

(define (cairo-graphics/close device)
  (let ((graphics (graphics-device/descriptor device)))
    (let ((surface (cairo-graphics-surface graphics)))
      (if surface
	  (let ((cr (cairo-graphics-context graphics)))
	    (cairo-destroy cr)
	    (set-cairo-graphics-context! graphics #f)
	    (cairo-surface-destroy surface)
	    (set-cairo-graphics-surface! graphics #f))))))

(define (cairo-graphics/device-coordinate-limits device)
  (let ((graphics (graphics-device/descriptor device)))
    (values 0 (cairo-graphics-height graphics)
	    (cairo-graphics-width graphics) 0)))

(define (cairo-graphics/coordinate-limits device)
  (apply values (cairo-graphics-limits (graphics-device/descriptor device))))

(define (cairo-graphics/set-coordinate-limits device x-left y-bottom x-right y-top)
  (let ((graphics (graphics-device/descriptor device)))
    (let ((cr (cairo-graphics-context graphics)))
      (if (zero? (- x-right x-left)) (error "Zero width coordinate limits:" 'x-left x-left 'y-bottom y-bottom 'x-right x-right 'y-top y-top))
      (if (zero? (- y-bottom y-top)) (error "Zero height coordinate limits:" 'x-left x-left 'y-bottom y-bottom 'x-right x-right 'y-top y-top))
      (cairo-restore cr)	;back to device coords.
      (cairo-save cr)
      (let ((x-factor (->flonum (/ (-1+ (cairo-graphics-width graphics))
				   (- x-right x-left))))
	    (y-factor (->flonum (/ (-1+ (cairo-graphics-height graphics))
				   (- y-bottom y-top)))))
	(cairo-scale cr x-factor y-factor)
	(let ((factor (flo:min (flo:abs x-factor) (flo:abs y-factor))))
	  (set-cairo-graphics-scale! graphics factor)
	  (cairo-set-line-width cr (flo:/ 1.0 factor))
	  (cairo-set-font-matrix cr (let ((k (flo:/ 10.0 factor)))
				      (cairo-matrix k 0. 0.
						    0. (flo:negate k) 0.)))))
      (cairo-translate cr (- x-left) (- y-top))
      (let ((fgcolor (cairo-graphics-fgcolor graphics)))
	(cairo-set-source-color cr fgcolor)))
    (set-cairo-graphics-limits! graphics (list x-left y-bottom x-right y-top))))

(define (cairo-graphics/clear device)
  (let ((graphics (graphics-device/descriptor device)))
    (let ((cr (cairo-graphics-context graphics))
	  (bgcolor (cairo-graphics-bgcolor graphics))
	  (bgcolor-name (cairo-graphics-bgcolor-name graphics)))
      (set-cairo-graphics-fgcolor! graphics bgcolor)
      (set-cairo-graphics-fgcolor-name! graphics bgcolor-name)
      (cairo-set-source-color cr bgcolor)
      (cairo-reset-clip cr)
      (cairo-paint cr))))

(define (cairo-graphics/draw-point device x y)
  (let ((graphics (graphics-device/descriptor device)))
    (let ((cr (cairo-graphics-context graphics))
	  (x (->flonum x))
	  (y (->flonum y))
	  (radius (/ 3.0 (cairo-graphics-scale device))))
      (cairo-arc cr x y radius 0. 2pi)
      (cairo-stroke cr))))

(define (cairo-graphics/draw-line device x-start y-start x-end y-end)
  (let ((graphics (graphics-device/descriptor device)))
    (let ((cr (cairo-graphics-context graphics))
	  (x (->flonum x-start))
	  (y (->flonum y-start)))
      (let ((dx (flo:- (->flonum x-end) x))
	    (dy (flo:- (->flonum y-end) y)))
	(cairo-move-to cr x y)
	(cairo-rel-line-to cr dx dy)
	(cairo-stroke cr)))))

(define (cairo-graphics/draw-text device x y string)
  (let ((graphics (graphics-device/descriptor device)))
    (let ((cr (cairo-graphics-context graphics))
	  (x (->flonum x))
	  (y (->flonum y)))
      (cairo-move-to cr x y)
      (cairo-show-text cr string))))

(define-integrable 2pi (flo:* 8. (flo:atan2 1. 1.)))

(define (cairo-graphics/draw-circle device x y radius)
  (let ((graphics (graphics-device/descriptor device)))
    (let ((cr (cairo-graphics-context graphics))
	  (x (->flonum x))
	  (y (->flonum y))
	  (radius (->flonum radius)))
      (cairo-arc cr x y radius 0. 2pi)
      (cairo-stroke cr))))

(define (cairo-graphics/move-cursor device x y)
  (let ((graphics (graphics-device/descriptor device)))
    (let ((cr (cairo-graphics-context graphics))
	  (x (->flonum x))
	  (y (->flonum y)))
      (cairo-move-to cr x y))))

(define (cairo-graphics/drag-cursor device x y)
  (let ((graphics (graphics-device/descriptor device)))
    (let ((cr (cairo-graphics-context graphics))
	  (x (->flonum x))
	  (y (->flonum y)))
      (cairo-line-to cr x y)
      (cairo-stroke cr))))

(define (cairo-graphics/set-drawing-mode device mode)
  (let ((operator
	 (case mode
	   ((0) 'clear)			;GXclear         0
	   ((1) 'in)			;GXand           src and dst
	   ((2) 'out)			;GXandReverse    src and not dst
	   ((3) 'source)		;GXcopy          src
	   ((4) 'dest-out)		;GXandInverted   not src and dst
	   ((5) 'dest)			;GXnoop          dst
	   ((6) 'xor)			;GXxor           src xor dst
	   ((7) 'over)			;GXor            src or dst
	   ((8) (warn "unimplemented:" '|GXnor|) #f)           ;GXnor           not src and not dst
	   ((9) (warn "unimplemented:" '|GXequiv|) #f)         ;GXequiv         not src xor dst
	   ((10) (warn "unimplemented:" '|GXinvert|) #f)       ;GXinvert        not dst
	   ((11) (warn "unimplemented:" '|GXorReverse|) #f)    ;GXorReverse     src or not dst
	   ((12) (warn "unimplemented:" '|GXcopyInverted|) #f) ;GXcopyInverted  not src
	   ((13) (warn "unimplemented:" '|GXorInverted|) #f)   ;GXorInverted    not src or dst
	   ((14) (warn "unimplemented:" '|GXnand|) #f)         ;GXnand          not src or not dst
	   ((15) 'source)		;GXset           1
	   (else (error:wrong-type-argument mode "a drawing mode"
					    'cairo-graphics/set-drawing-mode)))))
    (if operator
	(cairo-set-operator
	 (cairo-graphics-context (graphics-device/descriptor device))
	 operator))))

(define (cairo-graphics/set-line-style device style)
  (let ((graphics (graphics-device/descriptor device))
	(dashes
	 (case style
	   ((0) '())
	   ((1) '(8.))
	   ((2) '(1.))
	   ((3) '(13. 1. 1. 1.))
	   ((4) '(11. 1. 1. 1. 1. 1.))
	   ((5) '(11. 5.))
	   ((6) '(12. 1. 2. 1))
	   ((7) '( 9. 1. 2. 1. 2. 1.))
	   (else (error:wrong-type-argument style "a line style"
					    'cairo-graphics/set-line-style)))))
    (let ((cr (cairo-graphics-context graphics))
	  (factor (flo:/ 16. (cairo-graphics-scale graphics))))
      (cairo-set-dash cr (map (lambda (len) (flo:* len factor)) dashes)))))

(define (cairo-graphics/fill-polygon-list device points)
  ;; POINTS should be a list of flo:vectors each with a length greater than 1.
  (let ((graphics (graphics-device/descriptor device)))
    (let ((cr (cairo-graphics-context graphics))
	  (p (car points)))
      (define-integrable (x p) (flo:vector-ref p 0))
      (define-integrable (y p) (flo:vector-ref p 1))
      (cairo-move-to cr (x p) (y p))
      (for-each (lambda (p)
		  (cairo-line-to cr (x p) (y p)))
		(cdr points))
      (cairo-close-path cr)
      (cairo-fill cr))))

(define (cairo-graphics/flush device)
  (cairo-surface-flush
   (or (cairo-graphics-surface
	(graphics-device/descriptor device))
       (error "Cairo graphics device closed:" device))))

(define (cairo-graphics/write-to-png device filename)
  (cairo-surface-write-to-png
   (or (cairo-graphics-surface
	(graphics-device/descriptor device))
       (error "Cairo graphics device closed:" device))
   (->namestring filename)))

(define (cairo-graphics/destroy device)
  (cairo-surface-destroy (cairo-graphics-surface
			  (graphics-device/descriptor device))))

(define (cairo-graphics/set-background-color device name)
  (let ((graphics (graphics-device/descriptor device)))
    (if (not (string=? name (cairo-graphics-bgcolor-name graphics)))
	(let ((new (with-glib-lock
		    (lambda ()
		      (->color name 'cairo-graphics/set-background-color)))))
	  (set-cairo-graphics-bgcolor! graphics new)
	  (set-cairo-graphics-bgcolor-name! graphics name)))))

(define (cairo-graphics/set-foreground-color device name)
  (let ((graphics (graphics-device/descriptor device)))
    (if (not (string=? name (cairo-graphics-fgcolor-name graphics)))
	(let ((new (with-glib-lock
		    (lambda ()
		      (->color name 'cairo-graphics/set-foreground-color)))))
	  (set-cairo-graphics-fgcolor! graphics new)
	  (set-cairo-graphics-fgcolor-name! graphics name)
	  (cairo-set-source-color (cairo-graphics-context graphics) new)))))

(define (cairo-graphics/set-clip-rectangle device x-left y-bottom x-right y-top)
  (let ((cr (cairo-graphics-context (graphics-device/descriptor device)))
	(x (->flonum x-left))
	(y (->flonum y-top)))
    (let ((width (flo:- (->flonum x-right) x))
	  (height (flo:- (->flonum y-bottom) y)))
      (cairo-rectangle cr x y width height)
      (cairo-clip cr))))

(define (cairo-graphics/reset-clip-rectangle device)
  (cairo-reset-clip (cairo-graphics-context
		     (graphics-device/descriptor device))))

(define cairo-graphics-device-type)

(define (initialize-package!)
  (set! cairo-graphics-device-type
	(make-graphics-device-type
	 'cairo
	 `((available? ,cairo-graphics/available?)
	   (open ,cairo-graphics/open)
	   (clear ,cairo-graphics/clear)
	   (close ,cairo-graphics/close)
	   (coordinate-limits ,cairo-graphics/coordinate-limits)
	   (device-coordinate-limits ,cairo-graphics/device-coordinate-limits)
	   (move-cursor ,cairo-graphics/move-cursor)
	   (drag-cursor ,cairo-graphics/drag-cursor)
	   (draw-line ,cairo-graphics/draw-line)
	   (draw-point ,cairo-graphics/draw-point)
	   (draw-text ,cairo-graphics/draw-text)
	   (draw-circle ,cairo-graphics/draw-circle)
	   (flush ,cairo-graphics/flush)
	   (reset-clip-rectangle ,cairo-graphics/reset-clip-rectangle)
	   (set-clip-rectangle ,cairo-graphics/set-clip-rectangle)
	   (set-coordinate-limits ,cairo-graphics/set-coordinate-limits)
	   (set-drawing-mode ,cairo-graphics/set-drawing-mode)
	   (set-line-style ,cairo-graphics/set-line-style)))))

(initialize-package!)